/* -*- mode: c; c-basic-offset: 2; -*- */
/******** <lalVerbatim file="slopeC"> ********
Author: Daw, E
$Id: slope.c,v 1.2 2001/08/27 19:26:53 edaw Exp $
********* </lalVerbatim> ********/

#ifdef DEBUG_SLOPE
#include <stdio.h>
#endif
#include <math.h>
/* functions for memory allocation */
#include <lal/LALStdlib.h>
/* functions for vector creation and manipulation */
/* #include <lal/AVFactories.h> */
/* functions for LAL MPI interface */
#include <lal/Comm.h>
/* functions for slope filtering and peak searching data */
#include <lal/SlopeDetectorFilter.h>
/* structures and definitions for this wrapper code */
#include <Slope.h>

NRCSID( SLOPEC, "$Id: slope.c,v 1.2 2001/08/27 19:26:53 edaw Exp $");

/******** local static function declarations ********/

static
void
AddEventToLinkedList(LALStatus*        status,
		     INT4              eventTimeSecs,
                     INT4              eventTimeNSecs,
		     REAL4             eventAmp,
		     REAL4             eventDur,
		     REAL4             eventCF,
		     REAL4             eventBW,
		     SlopeDetections** pcurrentEvent);

static
void
MakeEventTable(LALStatus*            status,
	       SlopeSearchParams*    params,
	       LALSearchOutput*      output);

static
void
ApplySearchMaster(LALStatus*            status,
		  LALSearchOutput*      output,
		  SlopeSearchParams*    params);

/** note that currently my code is designed to run **/
/** on a single machine. ****************************/
/****************************************************/

/******** wrapperAPI function definitions ***********/

/* Initialize some parameters for the search and dynamically */
/* allocate some memory */

void LALInitSearch(LALStatus*           status,
		   void**               psearchParams,
		   LALInitSearchParams* initSearchParams)
{
  SlopeSearchParams* searchParams;

  INITSTATUS( status, "LALInitSearch", SLOPEC );
  ATTATCHSTATUSPTR(status);

  /* allocate memory for search parameter buffer */
  searchParams = LALMalloc(sizeof(SlopeSearchParams));
  *psearchParams = (void*) searchParams;

  /* set search started bit to false */
  searchParams->searchStarted = 0;

  /* initialize search parameters for simple search     */
  /* most of these values are just placeholders for now */
  strcpy(searchParams->channelName,
	 SLOPE_DEFAULT_CHANNEL);                    /*search channel*/

  searchParams->samplingRate = 
    SLOPE_DEFAULT_SAMPLING_RATE ;                   /*sampling rate*/

  searchParams->numberOfBinsForSlopeMeasure = 
    SLOPE_DEFAULT_MEASURE_WIDTH ;                   /*bins for slope measure*/

  searchParams->timeIntervalForSlopeMeasure =       /*interval for slope*/
    searchParams->numberOfBinsForSlopeMeasure /     /*measure */
    searchParams->samplingRate ;

  searchParams->peakWidth =                         /*number of adjacent*/
    SLOPE_DEFAULT_CORRELATED_WIDTH ;                /*filtered data bins*/
                                                    /*grouped under 1 peak*/
                                                    /*in peak search*/

  searchParams->triggerThreshold = 
    SLOPE_DEFAULT_THRESHOLD ;                       /*threshold for trigger*/

  searchParams->numberOfTriggersSoFar = 0;          /*counter for triggers*/

  searchParams->nSamplesInHistoryBuffer =           /*length of history */
    searchParams->numberOfBinsForSlopeMeasure +     /*buffer */
    searchParams->peakWidth - 2;      

  searchParams->numberOfInitialBinsToSkip = 0;      /*number of bins at the*/
                                                   /*start of a new data set*/
                                                    /*to skip, including */
                                                    /*history buffer*/

  searchParams->nData = 0;                          /*length of new data*/

  /* allocate memory for data history buffer for data */
  /* kept so that slope filter continuous at joins between */
  /* data stretches */
  searchParams->dataHistoryBuffer = 
    LALCalloc( searchParams->nSamplesInHistoryBuffer,
	       sizeof(REAL4) );

  /* initialize event list */
  searchParams->firstEvent = 0x0;
  searchParams->currentEvent = 0x0;

  /* initialize mpi information */
  searchParams->mpiMaster = initSearchParams->nodeClass;

  DETATCHSTATUSPTR(status);
  RETURN(status);
}

/* look in the data for the right channel, extract the */
/* channel data and any history buffer data to a vector */
/* of type REAL4. Get the start and end times of the data */
/* segment. */

void
LALConditionData(LALStatus*             status,
		 LALSearchInput*        inout,
		 void*                  searchParams,
		 LALMPIParams*          mpiParams)
{
  BOOLEAN dataFound = 0;
  UINT4 dCount = 0;
  UINT4 dElCount = 0;
  SlopeSearchParams* params;
  UINT4 dataArraySize = 0;
  REAL4 historyBufferTimeInterval;
  /* these variables needed for use with a math.h function */
  /* do we have a LAL datatype - safe equivalent of math.h:modf() ?*/
  double historyLengthSecs;
  double historyLengthNSecs;

  INITSTATUS( status, "LALConditionData", SLOPEC );
  ATTATCHSTATUSPTR(status);

  params = (SlopeSearchParams*) searchParams;

  /*loop through the data sets in the input object and*/
  /*get the data from the channel I requested*/
  for(dCount = 0; dCount < inout->numberSequences; ++dCount) {
    if( strcmp(inout->sequences[dCount].name,params->channelName) == 0 ) {
      /* error trap for non timeseries data here */
      if( inout->sequences[dCount].space != timeD ) {
	ABORT( status, CONDITIONDATA_NONTSDATA, CONDITIONDATA_NONTSDATAMESG);
      }
      dataFound = 1;
      dataArraySize = inout->sequences[dCount].dimensions[0] +
	params->nSamplesInHistoryBuffer;
      params->dataNew = LALCalloc( dataArraySize , sizeof(REAL4) );
      if( params->searchStarted == 0 ) {
	/* set history buffer elements to zeros */
	for( dElCount = 0 ; dElCount < (params->nSamplesInHistoryBuffer) ;
	     ++dElCount ) {
	  params->dataNew[dElCount] = 0;
	}
	/* tell search routine to ignore the history buffer */
	params->numberOfInitialBinsToSkip = 
	  params->nSamplesInHistoryBuffer;
	/* set search started bit to 1 */
	params->searchStarted = 1;
      } else {
	/* set first elements of data vector to history buffer */
	for( dElCount = 0 ; dElCount < (params->nSamplesInHistoryBuffer) ;
	     ++dElCount ) {
	  params->dataNew[dElCount] = params->dataHistoryBuffer[dElCount];
	}
      }	
      /* read new data to rest of data vector. The */
      /* cast to reals depends on the original data type. For now */
      /* I support only REAL4, UINT4 and UINT2 */
      switch ( inout->sequences[dCount].type ) {
      case int_2u:
	for( dElCount = params->nSamplesInHistoryBuffer ; dElCount <
	       dataArraySize ; ++dElCount ) {
	  params->dataNew[dElCount] = 
	    (REAL4) inout->sequences[dCount].data.int2u[dElCount] ;
	}
	break;
      case int_4u:
	for( dElCount = params->nSamplesInHistoryBuffer ; dElCount <
	       dataArraySize ; ++dElCount ) {
	  params->dataNew[dElCount] = 
	    (REAL4) inout->sequences[dCount].data.int4u[dElCount] ;
	}
	break;
      case real_4:
	for( dElCount = params->nSamplesInHistoryBuffer ; dElCount <
	       dataArraySize ; ++dElCount ) {
	  params->dataNew[dElCount] = 
	    inout->sequences[dCount].data.real4[dElCount] ;
	}
	break;
      default:
	ABORT( status, CONDITIONDATA_DATATYPEUNSUP,
	       CONDITIONDATA_DATATYPEUNSUPMESG );
      }

      /* set data array length */
      params->nData = dataArraySize;

      /* get data start time */
      historyBufferTimeInterval = params->nSamplesInHistoryBuffer /
	params->samplingRate;
      /* need to split history length into secs, nsecs. Will use */
      /* the modf function of math.h to do this, which necessitates */
      /* casts into machine-dependent data types. Sorry. */
      historyLengthNSecs = 1e+9 * modf((double)historyBufferTimeInterval,
				&historyLengthSecs);
      params->startTimeSec = 
	inout->sequences[dCount].range.dTime.startSec -
	(UINT4)historyLengthSecs;
      params->startTimeNan =
	inout->sequences[dCount].range.dTime.startNan -
	(UINT4)historyLengthNSecs;
    }
    /* break out of for loop once data located */
    if(dataFound == 1) {
      break;
    }
  }

  if( dataFound == 0 ) {
    /* error and exit, data channel was not present */
    ABORT( status, CONDITIONDATA_NOSUCHCHANNEL,
	   CONDITIONDATA_NOSUCHCHANNELMESG );
  }

  DETATCHSTATUSPTR(status);
  RETURN(status);
}

/* apply the slope filter to the data using the slopefilters */
/* lal package. Apply a peak search algorithm to the filtered */
/* data and add an entry to the event linked list for every */
/* peak that passes the cut. Increment the output data field */
/* that keeps track of the percentage of the analysis completed. */

void
LALApplySearch(LALStatus*                status,
	       LALSearchOutput*          output,
               LALSearchInput*           input,
	       LALApplySearchParams*     applyParams)
{
  SlopeSearchParams* params;

  INITSTATUS( status, "LALApplySearch", SLOPEC );
  ATTATCHSTATUSPTR(status);

  /* extract my parameters from the search parameters */
  params = (SlopeSearchParams*)applyParams->searchParams ;

  /* figure out your rank */
  MPI_Comm_rank(*(applyParams->mpiParams->comm) , &(params->mpiRank) );

  /* if you are the master, start looking for peaks. Otherwise */
  /* abort with an error because I currently don't support multiprocessing */
  if(params->mpiMaster) {
    ApplySearchMaster(status,output,params);
  } else {
    ABORT( status, APPLYSEARCH_NOSLAVES, APPLYSEARCH_NOSLAVESMESG );
  }

  /* write the linked list of located events to the */
  /* database */
  MakeEventTable(status, params, output);

  /* set done bit on output data */
  output->notFinished = 0;

  CHECKSTATUSPTR(status);
  DETATCHSTATUSPTR(status);
  RETURN(status);
}

/* free up memory dynamically assigned during analysis run */
/* note that I DO NOT free up memory allocated to the subfields */
/* of LALSearchOutput* output. There is no way to do this as it */
/* is not passed as an argument to LALFinalizeSearch. I am assuming */
/* that wrapperAPI tidies this up as necessary */

void
LALFinalizeSearch(LALStatus*             status,
		  void**                 psearchParams)
{
  SlopeSearchParams* searchParams = 0x0;
  SlopeDetections* pslope = 0x0;

  INITSTATUS( status, "LALFinalizeSearch", SLOPEC );
  ATTATCHSTATUSPTR(status);

  searchParams = (SlopeSearchParams*)*psearchParams;
  pslope = searchParams->firstEvent;
  while( (pslope->next) != 0x0 ) {
    pslope = pslope->next;
    LALFree(pslope->previous);
  }
  LALFree( pslope );
  LALFree( searchParams->dataNew );
  LALFree( searchParams->dataHistoryBuffer );
  LALFree( searchParams );

  DETATCHSTATUSPTR(status);
  RETURN(status);
}

/******** local static function definitions ********/

static
void
MakeEventTable(LALStatus* status,
		  SlopeSearchParams* slopeparams,
		  LALSearchOutput* output) {
  /* program name appearing in program field of SLOPE_FILTER table */

  /*this field used in Patrick Brady's moments code, not in power*/
  /*const CHAR* const PROGRAM = "slope";*/

  /* pointer to an output structure */
  outPut* poutput = 0x0;
  /* pointers to a dataBase structure */
  dataBase* db = 0x0;
  dataBase* dbprevious = 0x0;
  /* counter for events in the event list */
  UINT4 eventCounter = 0;
  /* pointer to an event list element structure */
  SlopeDetections* eventList;
  /* counter for colums in the database table */
  UINT4 dbCC=0;
 
  INITSTATUS( status, "MakeEventTable", SLOPEC );
  ATTATCHSTATUSPTR(status);

  /* set number of events to number of peaks found */
  output->numOutput = slopeparams->numberOfTriggersSoFar;
  /* allocate memory for database entries for each trigger */
  poutput = LALCalloc( output->numOutput , sizeof( outPut ) ); 
  output->result = poutput;

  /* loop over events, filling the outPut structure for each event */
  eventCounter = 0;
  eventList = slopeparams->firstEvent;
  while( eventCounter < (slopeparams->numberOfTriggersSoFar)) {

    (poutput+eventCounter)->search           = burst;
    (poutput+eventCounter)->significant      = 1;
    (poutput+eventCounter)->states           = NULL;
    (poutput+eventCounter)->optional         = NULL;
    (poutput+eventCounter)->indexNumber      = 0;
    (poutput+eventCounter)->results = LALMalloc( sizeof(dataBase) );
    db = (poutput+eventCounter)->results;
    /* code to make a single line of the database goes here */ 
    for(dbCC = 0; dbCC<7; ++dbCC) {
      if(dbCC == 0) {
	db->previous = NULL;
      } else {
	db->previous = dbprevious;
      }
      sprintf(db->tableName,"slope");
      db->numberRows = 1;
      db->rowDimensions = NULL;
      switch (dbCC) {
      case 0:
	sprintf(db->columnName,"event_id");
	db->type = int_4u;
	db->rows.int4u = (UINT4*) LALMalloc(sizeof(UINT4));
	*(db->rows.int4u) = eventCounter;
	break;
      case 1:
	sprintf(db->columnName,"start_time");
	db->type = int_4u;
	db->rows.int4u = (UINT4*) LALMalloc(sizeof(UINT4));
	*(db->rows.int4u) = eventList->eventDetectionTimeSeconds;
	break;
      case 2:
	sprintf(db->columnName,"start_time");
	db->type = int_4u;
	db->rows.int4u = (UINT4*) LALMalloc(sizeof(UINT4));
	*(db->rows.int4u) = eventList->eventDetectionTimeNanoseconds;
	break;
      case 3:
	sprintf(db->columnName,"duration");
	db->type = real_4;
	db->rows.real4 = (REAL4*) LALMalloc(sizeof(REAL4));
	*(db->rows.real4) = eventList->eventDuration;
	break;
      case 4:
	sprintf(db->columnName,"central_freq");
	db->type = real_4;
	db->rows.real4 = (REAL4*) LALMalloc(sizeof(REAL4));
	*(db->rows.real4) = eventList->eventCentralFrequency;
	break;
      case 5:
	sprintf(db->columnName,"bandwidth");
	db->type = real_4;
	db->rows.real4 = (REAL4*) LALMalloc(sizeof(REAL4));
	*(db->rows.real4) = eventList->eventBandwidth;
	break;
      case 6:
	sprintf(db->columnName,"snr");
	db->type = real_4;
	db->rows.real4 = (REAL4*) LALMalloc(sizeof(REAL4));
	*(db->rows.real4) = eventList->eventAmplitude;
	break;
      }
      if( dbCC == 6 ) {
	db->next = NULL;
      } else {
	db->next = LALMalloc( sizeof(dataBase) );
	dbprevious = db;
	db = db->next;
      }
    }
    ++eventCounter;
    eventList = (eventList->next);
  }

  DETATCHSTATUSPTR(status);
  RETURN(status);
}

static
void
AddEventToLinkedList(LALStatus* const  status,
		     INT4              eventTimeSecs,
		     INT4              eventTimeNSecs,
		     REAL4             eventAmp,
		     REAL4             eventDur,
		     REAL4             eventCF,
		     REAL4             eventBW,
		     SlopeDetections** pcurrentEvent) 
{
  SlopeDetections* newEvent;
  SlopeDetections* currentEvent = (*pcurrentEvent);

  /* allocate memory for new event. Write pointer to next field */
  /* of current event */
  currentEvent->next = LALMalloc( sizeof(SlopeDetections) );
  newEvent = currentEvent->next;

  /* fill in physical data for new event */
  newEvent->eventDetectionTimeSeconds = eventTimeSecs;
  newEvent->eventDetectionTimeNanoseconds = eventTimeNSecs;
  newEvent->eventAmplitude = eventAmp;
  newEvent->eventDuration = eventDur;
  newEvent->eventCentralFrequency = eventCF;
  newEvent->eventBandwidth = eventBW;

  /* write pointer to current event (the one passed to the function) */
  /* as the previous field of the new event */
  newEvent->previous = currentEvent;
  newEvent->next = 0x0;

  /* increment the current event to the new event */
  currentEvent = newEvent;

  RETURN(status);
}

static
void
ApplySearchMaster(LALStatus*          status,
		  LALSearchOutput*    output,
		  SlopeSearchParams*  params) 
{
  REAL4Vector dataIn;
  REAL4Vector dataSendOut;
  REAL4* dataOut;
  UINT4 dataSearchCount = 0;
  UINT4 nSearch = 0;
  UINT4 nDataOut = 0;
  UINT4 peakBin = 0;
  UINT4 binOffsetSec;
  UINT4 binOffsetNan;
  REAL4 peakAmplitude = 0.0;
  /* again, have to use math.h:modf() to break apart */
  /* a time in seconds stored as a REAL4 into two    */
  /* unsigned integers denoting seconds and nanoseconds*/
  /* this means I have to cast to double. Anyone have */
  /* a better way of doing this? */
  double nanoOut=0.0;
  double secOut=0;
  
  INITSTATUS( status, "ApplySearchMaster", SLOPEC );
  ATTATCHSTATUSPTR(status);

  /* get the length of the data output array */
  nDataOut = params->nData - params->numberOfBinsForSlopeMeasure + 1 ;

  /* get the number of bins to search */
  nSearch = params->nData - params->peakWidth - 1 ;

  /* construct data vector for input data array */
  dataIn.length = params->nData;
  dataIn.data = (REAL4*)params->dataNew;

  /* allocate memory for output data */
  dataOut = LALCalloc( nDataOut , sizeof(REAL4) );
  
  /* make structure containing output data array and size */
  dataSendOut.length = nDataOut;
  dataSendOut.data = dataOut;

  /* filter the data */
  LALSlopeDetectorFilter(status,
			 &dataSendOut,
			 &dataIn,
			 params->numberOfBinsForSlopeMeasure);

  /* do a peak search on the filtered data */
  dataSearchCount = params->numberOfInitialBinsToSkip;
  while(dataSearchCount < nSearch) {

    /* test if the next bin exceeds the threshold */
    if( dataOut[dataSearchCount] > params->triggerThreshold ) {
      ++params->numberOfTriggersSoFar;
      peakAmplitude = dataOut[dataSearchCount];
      peakBin = dataSearchCount;

      /* look over the (peakWidth) bins after the one exceeding */
      /* the threshold. The peak is positioned at the tallest one */
      while(dataSearchCount < ( peakBin + 
			     params->peakWidth - 1 )) {
	++dataSearchCount;
	if( dataOut[dataSearchCount] > peakAmplitude ) {
	  peakAmplitude = dataOut[dataSearchCount];
	  peakBin = dataSearchCount;
	}
      }

      /* get peak time */
      nanoOut = 1e+9 * modf(peakBin*(double)( 1 / params->samplingRate ),
			    &secOut);
      binOffsetNan = (UINT4)nanoOut;
      binOffsetSec = (UINT4)secOut;

      /* write newly located peak to linked list */
      AddEventToLinkedList(status,
			   (params->startTimeSec + binOffsetSec),
			   (params->startTimeNan + binOffsetNan),
			   peakAmplitude,
			   (params->peakWidth / params->samplingRate),
			   params->samplingRate / 4,
			   params->samplingRate / 2,
			   &(params->currentEvent) );
    } else {
      ++dataSearchCount;
    }
    /* update progress in search */
    output->fracRemaining = (REAL4)( (nSearch - dataSearchCount) / nSearch );
  }

  LALFree(dataOut);

  DETATCHSTATUSPTR(status);
  RETURN(status);
}




